/* ©  Copyright Steve Simon, 2013 */
#include <u.h>
#include <libc.h>
#include <ctype.h>
#include <auth.h>
#include <fcall.h>
#include <thread.h>
#include <9p.h>
#include <json.h>
#include "jira.h"

Node *
uservalid(Node *n, Sess *s, char *user)
{
	Node *cc;
	Jvalue *v;

	if((v = get(s, "rest/api/%s/user?username=%s&", s->api, user)) == nil)
		return nil;
	Jfree(v);
	cc = subdir(&n->child, nil, nil, user);

	scanhook(subdir(&cc, user, nil, user), userinfo, user);

	return cc;
}


/*
 * Caller may not have sufficent priviledge to see this info
 */
static void
security(Node **np, Sess *s, char *user)
{
	int i;
	Fmt f;
	Jvalue *v;
	char *p, *buf;

	if((v = get(s, "rest/api/%s/securitylevel/%s", user)) == nil){
		subtext(np, "securitylevel", "");
		(*np)->dir.mode &= ~0777;
		return;
	}

	fmtstrinit(&f);
	fmtprint(&f, "%-24q %s\n", "#Name", "Desc");
	for(i = 0; (p = Jstrval(v, "name", i)) != nil; i++){
		fmtprint(&f, "%-24q ", p);
		if((p = Jstrval(v, "description", i)) != nil)
			fmtprint(&f, "%s", p);
		fmtprint(&f, "\n");
	}
	if((buf = fmtstrflush(&f)) == nil)
		sysfatal("No memory\n");
	subtext(np, "securitylevel", buf);
	free(buf);
}

Node *
userinfo(Node *n, Sess *s, char *user)
{
	int i;
	Fmt f;
	Jvalue *v;
	char *p, *buf;

	if((v = get(s, "rest/api/%s/user?username=%s&expand=groups", s->api, user)) == nil){
		warn("%s cannot get user detail - %r\n", user);
		return n;
	}
	if((p = Jstrval(v, "displayName")) != nil)
		subtext(&n, "realname", p);
	if((p = Jstrval(v, "emailAddress")) != nil)
		subtext(&n, "email", p);
	if((p = Jstrval(v, "active")) != nil)
		subtext(&n, "active", p);
	if((p = Jstrval(v, "timeZone")) != nil)
		subtext(&n, "timezone", p);

	fmtstrinit(&f);
	for(i = 0; (p = Jstrval(v, "groups.items[%d].name", i)) != nil; i++){
		fmtprint(&f, "%s\n", p);
		addgroup(s, p);
	}
	if((buf = fmtstrflush(&f)) == nil)
		sysfatal("No memory\n");
	subtext(&n, "groups", buf);
	free(buf);

	security(&n, s, user);

	scanhook(subdir(&n, nil, nil, "queue"), queuelist, user);
	scanhook(subdir(&n, nil, nil, "reported"), reportedlist, user);
	scanhook(subdir(&n, nil, nil, "resolved"), resolvedlist, user);

	Jfree(v);
	return n;
}

/*
 * The Jira rest api has no way to enumerate the usernames known to jira
 * so we fake it by adding a /users dirrectory each username as we find
 * them.
 */
void
adduser(Sess *s, char *uid)
{
	Node *n;

	for(n = *(s->users); n; n = n->next)
		if(strcmp(n->dir.name, uid) == 0)
			return;
	scanhook(subdir(s->users, uid, nil, uid), userinfo, uid);
}

Node *
groupinfo(Node *n, Sess *s, char *group)
{
	Fmt f;
	Jvalue *v;
	int i, start, want;
	char *fmt, *user, *buf;

	start = 0;
	want = 1000;
	fmtstrinit(&f);
	fmt = "rest/api/%s/group?groupname=%s&expand=users[%d:%d]";
	do{
		if((v = get(s, fmt, s->api, group, want, start)) == nil){
			warn("%s cannot get group detail - %r\n", group);
			break;
		}
		for(i = 0; (user = Jstrval(v, "users.items[%d].name", i)) != nil; i++){
			adduser(s, user);
			fmtprint(&f, "%s\n", user);
		}
		Jfree(v);
		start += i;
	}while(i == want);

	if((buf = fmtstrflush(&f)) == nil)
		sysfatal("No memory\n");
	subtext(&n, "members", buf);
	free(buf);

	return n;
}

Node *
groupvalid(Node *n, Sess *s, char *group)
{
	Node *cc;
	Jvalue *v;

	if((v = get(s, "rest/api/%s/group?%s", s->api, group)) == nil)
		return nil;
	Jfree(v);
	cc = subdir(&n->child, nil, group, group);

	scanhook(subdir(&cc, nil, group, group), groupinfo, group);

	return cc;
}

void
addgroup(Sess *s, char *uid)
{
	Node *n;

	for(n = *(s->groups); n; n = n->next)
		if(strcmp(n->dir.name, uid) == 0)
			return;
	scanhook(subdir(s->groups, uid, nil, uid), groupinfo, uid);
}

Node *
issuevalid(Node *n, Sess *s, char *issue)
{
	Node *cc;
	Jvalue *v;
	char *lwr;

	if((v = get(s, "rest/api/%s/issue/%U/transitions", s->api, issue)) == nil)
		return nil;
	Jfree(v);

	lwr = strlwr(estrdup9p(issue));
	cc = subdir(&n->child, lwr, nil, issue);
	scanhook(subdir(&cc, lwr, nil, lwr), s->issueinfo, issue);
	free(lwr);

	return cc;
}

Node *
issuelist(Node *n, Sess *s, char *proj)		/* v5 compat */
{
	Jvalue *v;
	int i, want, start;
	char *fmt, *issue, *lwr;

	start = 0;
	want = 1000;
	fmt = "rest/api/%s/search?jql=project=%U&maxResults=%d&startAt=%d";
	do{
		if((v = get(s, fmt, s->api, proj, want, start)) == nil){
			warn("%s cannot get issue list - %r\n", proj);
			break;
		}
		for(i = 0; (issue = Jstrval(v, "issues[%d].key", i)) != nil; i++){
			lwr = strlwr(estrdup9p(issue));
			scanhook(subdir(&n, lwr, nil, lwr), s->issueinfo, issue);
			free(lwr);
		}

		Jfree(v);
		start += i;
	}while(i == want);
	return n;
}

static void
issuetypes(Node **np, Jvalue *v)
{
	int i;
	Fmt f;
	char *p, *buf;

	fmtstrinit(&f);
	fmtprint(&f, "%-24q %-12s %s\n", "#Name", "Subtask", "Desc");
	for(i = 0; (p = Jstrval(v, "issueTypes[%d].name", i)) != nil; i++){
		fmtprint(&f, "%-24q ", p);
		if((p = Jstrval(v, "issueTypes[%d].subtask", i)) != nil)
			fmtprint(&f, "%-12q", p);
		else
			fmtprint(&f, "%-12s", "-");
		if((p = Jstrval(v, "issueTypes[%d].description", i)) != nil)
			fmtprint(&f, "%s", p);
		fmtprint(&f, "\n");
	}
	if((buf = fmtstrflush(&f)) == nil)
		sysfatal("No memory\n");
	subtext(np, "issuetypes", buf);
	free(buf);
}

static void
versions(Node **np, Jvalue *v)		/* FIXME: Not checked for v5 compat yet */
{
	int i;
	Fmt f;
	char *p, *buf;

	fmtstrinit(&f);
	fmtprint(&f, "%-16s %-10s %-10s %s\n", "#Name", "Archived", "Released", "Description");
	for(i = 0; (p = Jstrval(v, "versions[%d].name", i)) != nil; i++){
		fmtprint(&f, "%-16q ", p);

		if((p = Jstrval(v, "versions[%d].archived", i)) != nil)
			fmtprint(&f, "%-10q ", p);
		else
			fmtprint(&f, "%-10s ", "-");
		if((p = Jstrval(v, "versions[%d].released", i)) != nil)
			fmtprint(&f, "%-10q ", p);
		else
			fmtprint(&f, "%-10s ", "-");
		if((p = Jstrval(v, "versions[%d].description", i)) != nil)
			fmtprint(&f, "%s\n", p);
		else
			fmtprint(&f, "%s\n", "-");
	}
	if((buf = fmtstrflush(&f)) == nil)
		sysfatal("No memory\n");
	if(strlen(buf) > 0)
		subtext(np, "versions", buf);
	free(buf);
}


static void
components(Node **np, Jvalue *v)
{
	int i;
	Fmt f;
	char *p, *buf;

	fmtstrinit(&f);
	fmtprint(&f, "%-32s %-16s %-16s %s\n", "#Name", "Lead", "Assignee", "Description");
	for(i = 0; (p = Jstrval(v, "components[%d].name", i)) != nil; i++){
		fmtprint(&f, "%-32q ", p);

		if((p = Jstrval(v, "components[%d].lead.name", i)) != nil)
			fmtprint(&f, "%-16s ", p);
		else
			fmtprint(&f, "%-16s ", "-");

		if((p = Jstrval(v, "components[%d].assignee.name", i)) != nil)
			fmtprint(&f, "%-16s ", p);
		else
			fmtprint(&f, "%-16s ", "-");

		if((p = Jstrval(v, "components[%d].description", i)) != nil)
			fmtprint(&f, "%-16s\n", p);
		else
			fmtprint(&f, "%-16s\n", "-");
	}
	if((buf = fmtstrflush(&f)) == nil)
		sysfatal("No memory\n");
	if(strlen(buf) > 0)
		subtext(np, "components", buf);
	free(buf);
}

Node *
projvalid(Node *n, Sess *s, char *proj)
{
	Node *cc, *cc2;
	Jvalue *v;
	char *lwr;

	if((v = get(s, "rest/api/%s/project/%U/versions", s->api, proj)) == nil)
		return nil;
	Jfree(v);

	cc = subdir(&n->child, nil, nil, proj);

	lwr = strlwr(estrdup9p(proj));
	cc2 = subdir(&cc, lwr, nil, lwr);
	scanhook(cc2, projinfo, proj);

	cc2 = subdir(&cc->child, nil, nil, "issues");		/* Temporary bodge until "walk not exist, use scan for walk" code is done */
	scanhook(cc2, issuelist, proj);
	walkhook(cc2, issuevalid, proj);

	free(lwr);
	return cc;
}

/*
 * only appears in V5 and later API but if the request
 * is rejected then we just don't create the file.
 *
 * only works if the current user has permission to
 * perform issue reassignment in this project, this is to
 * prevent email address harvesting by spammers.
 */
static void
assignable(Node **np, Sess *s, char *proj)
{
	Fmt f;
	Jvalue *v;
	char *p, *fmt, *buf;
	int i, start, want;

	start = 0;
	want = 1000;
	fmtstrinit(&f);
	fmt = "/rest/api/%s/user/assignable/search?project=%U&maxResults=%d&startAt=%d";
	do{
		if((v = get(s, fmt, s->api, proj, want, start)) == nil){
			subtext(np, "assignable", "");
			(*np)->dir.mode &= ~0777;
			return;
		}

		for(i = 0; i < want && (p = Jstrval(v, "[%d].name", i)) != nil; i++){
			fmtprint(&f, "%s\n", p);
			adduser(s, p);
		}

		Jfree(v);
		start += i;
	}while(i == want);

	if((buf = fmtstrflush(&f)) == nil)
		sysfatal("No memory\n");
	if(strlen(buf) > 0)
		subtext(np, "assignable", buf);
	free(buf);
}

/* you may not have permission to do this */
Node *
projinfo(Node *n, Sess *s, char *proj)
{
	Node *cc;
	Jvalue *v;
	char *p, *uid;

	if((v = get(s, "rest/api/%s/project/%U", s->api, proj)) == nil){
		warn("%s cannot get project detail - %r\n", proj);
		return n;
	}

	if((uid = Jstrval(v, "lead.name")) != nil){
		fulltext(&n, "lead", uid, nil, uid, 0, 0, uid);
		adduser(s, uid);
	}
	if((p = Jstrval(v, "name")) != nil)
		fulltext(&n, "title", uid, nil, uid, 0, 0, p);
	if((p = Jstrval(v, "description")) != nil)
		fulltext(&n, "description", uid, nil, uid, 0, 0, p);

	issuetypes(&n, v);
	versions(&n, v);
	components(&n, v);

	cc = subdir(&n, uid, nil, "issues");
	scanhook(cc, issuelist, proj);
	walkhook(cc, issuevalid, proj);

	assignable(&n, s, proj);

	Jfree(v);
	return n;
}

Node *
queuelist(Node *n, Sess *s, char *user)
{
	Jvalue *v;
	int i, start, want;
	char *issue, *fmt, *lwr;

	start = 0;
	want = 1000;
	fmt = "rest/api/%s/search?jql=assignee=%s+and+status!=Resolved&maxResults=%d&startAt=%d";
	do{
		if((v = get(s, fmt, s->api, user, want, start)) == nil){
			warn("%s cannot get user's queue - %r\n", user);
			break;
		}
		for(i = 0; (issue = Jstrval(v, "issues[%d].key", i)) != nil; i++){
			lwr = strlwr(estrdup9p(issue));
			scanhook(subdir(&n, nil, nil, lwr), s->issueinfo, issue);
			free(lwr);
		}
		Jfree(v);
		start += i;
	}while(i == want);
	return n;
}

Node *
resolvedlist(Node *n, Sess *s, char *user)
{
	Jvalue *v;
	int i, start, want;
	char *issue, *fmt, *lwr;

	start = 0;
	want = 1000;
	fmt = "rest/api/%s/search?jql=assignee=%s+and+status=Resolved&maxResults=%d&startAt=%d";
	do{
		if((v = get(s, fmt, s->api, user, want, start)) == nil){
			warn("%s cannot get user's queue - %r\n", user);
			break;
		}
		for(i = 0; (issue = Jstrval(v, "issues[%d].key", i)) != nil; i++){
			lwr = strlwr(estrdup9p(issue));
			scanhook(subdir(&n, nil, nil, lwr), s->issueinfo, issue);
			free(lwr);
		}
		Jfree(v);
		start += i;
	}while(i == want);
	return n;
}

Node *
reportedlist(Node *n, Sess *s, char *user)
{
	Jvalue *v;
	int i, start, want;
	char *issue, *fmt, *lwr;

	start = 0;
	want = 1000;
	fmt = "rest/api/%s/search?jql=reporter=%s&maxResults=%d&startAt=%d";
	do{
		if((v = get(s, fmt, s->api, user, want, start)) == nil){
			warn("%s cannot get user's queue - %r\n", user);
			break;
		}
		for(i = 0; (issue = Jstrval(v, "issues[%d].key", i)) != nil; i++){
			lwr = strlwr(estrdup9p(issue));
			scanhook(subdir(&n, nil, nil, lwr), s->issueinfo, issue);
			free(lwr);
		}
		Jfree(v);
		start += i;
	}while(i == want);
	return n;
}



Node *
projlist(Node *n, Sess *s, char *)
{
	int i;
	Jvalue *v;
	char *proj, *lwr;

	if((v = get(s, "rest/api/%s/project", s->api)) == nil){
		warn("cannot get project list - %r\n");
		return n;
	}

	for(i = 0; (proj = Jstrval(v, "[%d].key", i)) != nil; i++){
		lwr = strlwr(estrdup9p(proj));
		scanhook(subdir(&n, lwr, nil, lwr), projinfo, proj);
		free(lwr);
	}
	Jfree(v);
	return n;
}

Node *
serverinfo(Node *n, Sess *s, char *)
{
	char *p;
	Jvalue *v;

	if((v = get(s, "rest/api/%s/serverInfo", s->api)) == nil){
		warn("cannot get server detail - %r\n");
		return n;
	}


	if((p = Jstrval(v, "version")) != nil){
		subtext(&n, "version", p);
		s->major = strtol(p, nil, 10);
	}
	if((p = Jstrval(v, "buildNumber")) != nil)
		subtext(&n, "build-number", p);
	if((p = Jstrval(v, "buildDate")) != nil)
		subtext(&n, "build-time", ctime(iso2sec(p)));
	if((p = Jstrval(v, "buildPartnerName")) != nil)
		subtext(&n, "build-partner", p);
	if((p = Jstrval(v, "serverTime")) != nil)
		subtext(&n, "server-time", ctime(iso2sec(p)));
	if((p = Jstrval(v, "serverTitle")) != nil)
		subtext(&n, "title", p);
	if((p = Jstrval(v, "scmInfo")) != nil)
		subtext(&n, "scm-info", p);
	if((p = Jstrval(v, "baseUrl")) != nil)
		subtext(&n, "base-url", p);
	else
		subtext(&n, "base-url", s->url);
	Jfree(v);
	return n;
}

Node *
sessioninfo(Node *n, Sess *s, char *)
{
	char *p;
	Jvalue *v;

	/* This may be overwritten depending on server setup */
	if(s->annon)
		subtext(&n, "username", "<annonymous>");

	if((v = get(s, "rest/auth/1/session")) == nil){
		warn("cannot get session detail - %r\n");
		return n;
	}
	if((p = Jstrval(v, "name")) != nil){
		adduser(s, p);
		free(s->user);
		s->user = estrdup9p(p);
		fulltext(&n, "username", s->user, nil, s->user, 0, 0, s->user);
	}
	if((p = Jstrval(v, "loginInfo.loginCount")) != nil)
		fulltext(&n, "logins", s->user, nil, s->user, 0, 0, p);
	if((p = Jstrval(v, "loginInfo.previousLoginTime")) != nil)
		fulltext(&n, "last-login", s->user, nil, s->user, 0, 0, ctime(iso2sec(p)));
	if((p = Jstrval(v, "loginInfo.failedLoginCount")) != nil)
		fulltext(&n, "failed-logins", s->user, nil, s->user, 0, 0, p);
	if((p = Jstrval(v, "loginInfo.lastFailedLoginTime")) != nil)
		fulltext(&n, "last-failed-login", s->user, nil, s->user, 0, 0, ctime(iso2sec(p)));
	Jfree(v);
	return n;
}

static Node *
resolutions(Node *n, Sess *s, char *)
{
	int i;
	Fmt f;
	Jvalue *v;
	char *p, *buf;

	if((v = get(s, "/rest/api/%s/resolution", s->api)) == nil)
		return n;

	fmtstrinit(&f);
	fmtprint(&f, "%-24q %s\n", "#Name", "Description");
	for(i = 0; (p = Jstrval(v, "[%d].name", i)) != nil; i++){
		fmtprint(&f, "%-15q ", p);
		if((p = Jstrval(v, "[%d].description", i)) != nil)
			fmtprint(&f, "%s ", p);
		fmtprint(&f, "\n");
	}
	Jfree(v);

	if((buf = fmtstrflush(&f)) == nil)
		sysfatal("No memory\n");
	if(strlen(buf) > 0)
		subtext(&n, "resolutions", buf);
	free(buf);
	return n;
}

static Node *
statuses(Node *n, Sess *s, char *)
{
	int i;
	Fmt f;
	Jvalue *v;
	char *p, *buf;

	if((v = get(s, "/rest/api/%s/status", s->api)) == nil){
		warn("cannot get statuses - %r\n");
		return n;
	}

	fmtstrinit(&f);
	fmtprint(&f, "%-24q %s\n", "#Name", "Description");
	for(i = 0; (p = Jstrval(v, "[%d].name", i)) != nil; i++){
		fmtprint(&f, "%-24q ", p);
		if((p = Jstrval(v, "[%d].description", i)) != nil)
			fmtprint(&f, "%s ", p);
		fmtprint(&f, "\n");
	}
	Jfree(v);

	if((buf = fmtstrflush(&f)) == nil)
		sysfatal("No memory\n");
	if(strlen(buf) > 0)
		subtext(&n, "statuses", buf);
	free(buf);
	return n;
}

static Node *
linktypes(Node *n, Sess *s, char *)
{
	int i;
	Fmt f;
	Jvalue *v;
	char *p, *buf;

	if((v = get(s, "/rest/api/%s/issueLinkType", s->api)) == nil){
		warn("cannot get linktypes - %r\n");
		return n;
	}

	fmtstrinit(&f);
	fmtprint(&f, "%-32q %-16s %s\n", "#Name", "Inward", "Outward");
	for(i = 0; (p = Jstrval(v, "issueLinkTypes[%d].name", i)) != nil; i++){
		fmtprint(&f, "%-32q ", p);
		if((p = Jstrval(v, "issueLinkTypes[%d].inward", i)) != nil)
			fmtprint(&f, "%-16q ", p);
		if((p = Jstrval(v, "issueLinkTypes[%d].outward", i)) != nil)
			fmtprint(&f, "%q ", p);
		fmtprint(&f, "\n");
	}
	Jfree(v);

	if((buf = fmtstrflush(&f)) == nil)
		sysfatal("No memory\n");
	if(strlen(buf) > 0)
		subtext(&n, "linktypes", buf);
	free(buf);
	return n;
}

Node *
globalinfo(Node *n, Sess *s, char *arg)
{
	n = resolutions(n, s, arg);
	n = statuses(n, s, arg);
	n = linktypes(n, s, arg);
	return n;
}
